/*
 *  Amiga Linux/m68k Loader
 *
 *   Copyright 1995 by Geert Uytterhoeven
 *
 *
 *  The original idea for this program comes from the Linux distribution:
 *
 *	LILO - Generic Boot Loader for Linux ("LInux LOader")
 *	       by Werner Almesberger
 *
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING for more details.
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <elf.h>
#include <linux/major.h>
#include <linux/hdreg.h>

#include "lilo_proto.h"


const char LiloVersion[] = VERSION;

#ifndef BLKGETSIZE
#define BLKGETSIZE	_IO(0x12,96)
#endif
#ifndef BLKGETSIZE64
#define BLKGETSIZE64	_IOR(0x12,114,size_t)
#endif
#ifndef BLKSSZGET
#define BLKSSZGET	_IO(0x12,104)
#endif
#ifndef FIBMAP
#define FIBMAP		_IO(0x00,1)
#endif
#ifndef FIGETBSZ
#define FIGETBSZ	_IO(0x00,2)
#endif


    /*
     *	Command Line Options
     */

static const char *ProgramName = NULL;
static int Verbose = 0;
static int Install = 0;
static int DoBackup = 0;
static int Uninstall = 0;
static int Force = 0;
static int DoChecksum = 0;
static int DoCreateReset = 0;
static const char *DosType = NULL;
static const char *SaveBootBlock = NULL;
static const char *RestoreBootBlock = NULL;
static const char *Device = NULL;
static const char *Root = "";


    /*
     *	Boot Code Definitions
     */

extern u_char _BootCodeStart[];
extern u_char _BootCodeEnd[];
extern u_char _LoaderVectorStart[];
extern u_char _LoaderVectorEnd[];
extern u_char _AltDeviceName[];
extern u_char _AltDeviceUnit[];


    /*
     *	Reset Code Definitions
     */

extern u_char _ResetStart[];
extern u_char _ResetEnd[];


    /*
     *  Loader Code Definitions
     */

extern u_char _LoaderCodeStart[];
extern u_char _LoaderCodeEnd[];
extern struct DeviceDef _MapDevice;
extern u_char _MapDeviceName[];
extern u_char _MapExtentData[];

    /*
     *	Valid DosTypes
     */

struct DosTypeEntry {
    u_long ID;
    const char *ASCII;
    const char *Description;
};

static struct DosTypeEntry DosTypeEntries[] = {
    { ID_BOOT_DISK, "BOOU", "Generic boot disk" },
    { ID_OFS_DISK, "DOS0", "OFS disk" },
    { ID_FFS_DISK, "DOS1", "FFS disk" },
    { ID_OFS_INTL_DISK, "DOS2", "OFS INTL disk" },
    { ID_FFS_INTL_DISK, "DOS3", "FFS INTL disk" },
    { ID_OFS_DC_DISK, "DOS4", "OFS DC disk" },
    { ID_FFS_DC_DISK, "DOS5", "FFS DC disk" },
    { ID_MUFS_DISK, "muFS", "MultiUser FFS INTL disk" },
    { ID_MU_OFS_DISK, "muF0", "MultiUser OFS disk" },
    { ID_MU_FFS_DISK, "muF1", "MultiUser FFS disk" },
    { ID_MU_OFS_INTL_DISK, "muF2", "MultiUser OFS INTL disk" },
    { ID_MU_FFS_INTL_DISK, "muF3", "MultiUser FFS INTL disk" },
    { ID_MU_OFS_DC_DISK, "muF4", "MultiUser OFS DC disk" },
    { ID_MU_FFS_DC_DISK, "muF5", "MultiUser FFS DC disk" },
    { 0, NULL, NULL }
};


    /*
     *	Lilo Data
     */

static struct BootBlock BootBlock;

static int LoaderSize;
static u_long DosTypeID;

static struct DeviceDef *BootDevice, *AltDevice;


    /*
     *	Files
     */

const char *ConfigFile;
static struct FileDef *LoaderFile;
static struct FileDef *MapFile;
static const char *BackupFile;


    /*
     *	Function Prototypes
     */

int main(int argc, char *argv[]);

static void Error_Open(const char *name);
static void Error_OpenDir(const char *name);
static void Error_Read(const char *name);
static void Error_Write(const char *name);
static void Error_Stat(const char *name);
static void Error_Ioctl(const char *name, const char *ioctl);
static void WriteFile(const char *name, const void *data, int size);
static void ReadBootBlock(const char *name);
static void WriteBootBlock(const char *name);
static void CreateBootBlock(int reset);
static void BackupBootBlock(void);
static void SetDosType(void);
static void FixChecksum(void);
static u_long CalcChecksum(void);
static void ReadConfigFile(void);
static int EqualStrings(const char *s1, const char *s2);
static void FixSpecialChars(char *s);
static void CreateMapFile(void);
static void WriteTagData(int fd, u_long tag, const void *data, u_long size);
static void WriteTagString(int fd, u_long tag, const char *s);
static int CreateVector(struct FileDef *file);
static void Geometry(struct DeviceDef *device);
static void PatchLoader(void);
static void WriteLoader(void);
static const char *CreateFileName(const char *name);
static const char *CreateBackupFileName(void);
static void Usage(void) __attribute__ ((noreturn));


    /*
     *	Print an Error Message and Exit
     */

void Die(const char *fmt, ...)
{
    va_list ap;

    fflush(stdout);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}


    /*
     *  Error Conditions
     */

void Error_NoMemory(void)
{
    Die("Can't allocate memory\n");
}

static void Error_Open(const char *name)
{
    Die("Can't open file `%s': %s\n", name, strerror(errno));
}

static void Error_OpenDir(const char *name)
{
    Die("Can't open directory `%s': %s\n", name, strerror(errno));
}

static void Error_Read(const char *name)
{
    Die("Can't read from file `%s': %s\n", name, strerror(errno));
}

static void Error_Write(const char *name)
{
    Die("Can't write to file `%s': %s\n", name, strerror(errno));
}

static void Error_Stat(const char *name)
{
    Die("Can't get info about file `%s': %s\n", name, strerror(errno));
}

static void Error_Ioctl(const char *name, const char *ioctl)
{
    Die("Operation `%s' failed for file `%s': %s\n", ioctl, name,
	strerror(errno));
}

void *Malloc(size_t size)
{
    void *ptr = malloc(size);
    if (!ptr)
	Die("Can't allocate memory (%u bytes)\n", size);
    memset(ptr, 0, size);
    return ptr;
}


    /*
     *	Write a File
     */

static void WriteFile(const char *name, const void *data, int size)
{
    int fd;

    if (Verbose)
	printf("Writing file `%s'\n", name);

    if ((fd = open(name, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR)) == -1)
	Error_Open(name);
    if (write(fd, data, size) != size)
	Error_Write(name);
    close(fd);
}


    /*
     *	Read a Boot Block from a Device or File
     */

static void ReadBootBlock(const char *name)
{
    int fd;

    if (Verbose)
	printf("Reading boot block from `%s'\n", name);

    if ((fd = open(name, O_RDONLY)) == -1)
	Error_Open(name);
    if (read(fd, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
	Error_Read(name);
    close(fd);
}


    /*
     *	Write a Boot Block to a Device
     */

static void WriteBootBlock(const char *name)
{
    int fd;

    if (Verbose)
	printf("Writing boot block to device `%s'\n", name);

    if ((fd = open(name, O_RDWR)) == -1)
	Error_Open(name);
    if (write(fd, &BootBlock, sizeof(BootBlock)) != sizeof(BootBlock))
	Error_Write(name);
    close(fd);
}


    /*
     *	Create the Boot Block
     */

static void CreateBootBlock(int reset)
{
    u_long start, size, maxsize, data;

    if (Verbose)
	puts("Creating boot block");

    if (reset) {
	start = (u_long)_ResetStart;
	size = (u_long)_ResetEnd;
    } else {
	start = (u_long)_BootCodeStart;
	size = (u_long)_BootCodeEnd;
    }
    size -= start;
    data = (u_long)BootBlock.Data;
    maxsize = sizeof(BootBlock.Data);
    memset((void *)data, 0, maxsize);
    if (size > maxsize)
	Die("%ld bytes too much boot code\n", size - maxsize);
    if (Verbose)
	printf("%lu bytes boot code, %lu bytes free\n", size, maxsize - size);
    BootBlock.ID = DosTypeID;
    if (BootBlock.ID == ID_BOOT_DISK)
	BootBlock.DosBlock = 0;
    memcpy((void *)data, (void *)start, size);

    if (!reset) {
	maxsize = _LoaderVectorEnd - _LoaderVectorStart;
	if (LoaderFile->ExtentSize > maxsize)
	    Die("Loader vector is too large for boot block (%ld bytes too much)\n",
		LoaderFile->ExtentSize - maxsize);
	if (Verbose)
	    printf("%d bytes for loader vector blocks, %lu bytes free\n",
		   LoaderFile->ExtentSize, maxsize - LoaderFile->ExtentSize);
	memcpy((void *)(data + (u_long)_LoaderVectorStart - start), LoaderFile->Extent,
	       LoaderFile->ExtentSize);
	if (LoaderFile->device->parent != BootDevice->parent) {
	    struct DeviceDef *device = LoaderFile->device->parent;
	    strcpy((void *)(data + (u_long)_AltDeviceName - start), device->device);
	    *(u_int32_t *)(data + (u_long)_AltDeviceUnit - start) = device->unit;
	}
    }
    BootBlock.LiloID = LILO_ID;
    FixChecksum();
}


    /*
     *  Backup the Boot Block if it wasn't backed up before
     */

static void BackupBootBlock(void)
{
    struct stat info;

    if (stat(BackupFile, &info) == -1 || DoBackup) {
	WriteFile(BackupFile, &BootBlock, sizeof(BootBlock));
	printf("Boot block backed up to file `%s'\n", BackupFile);
    } else if (Verbose)
        printf("Boot block backup `%s' already exists. No backup taken.\n",
	       BackupFile);
}


    /*
     *	Set the DosType of the Boot Block
     */

static void SetDosType(void)
{
    int i;

    for (i = 0; DosTypeEntries[i].ID; i++)
	if (!strcmp(DosType, DosTypeEntries[i].ASCII))
	    break;
    if (DosTypeEntries[i].ID)
	DosTypeID = DosTypeEntries[i].ID;
    else {
	printf("Unknown DosType `%s'\n", DosType);
	puts("Valid DosTypes are:");
	for (i = 0; DosTypeEntries[i].ID; i++)
	    printf(" %s", DosTypeEntries[i].ASCII);
	Die("\n");
    }
    BootBlock.ID = DosTypeID;
    if (Verbose)
	printf("DosType set to `%s'\n", DosType);
    FixChecksum();
}


    /*
     *	Fix the Checksum of the Boot Block
     */

static void FixChecksum(void)
{
    if (Verbose)
	puts("Calculating boot block checksum");

    BootBlock.Checksum = 0;
    BootBlock.Checksum = CalcChecksum();
}


    /*
     *	Calculate the Boot Block Checksum
     */

static u_long CalcChecksum(void)
{
    u_long res;
    struct BootBlock *bb = &BootBlock;

#if __mc68000__
    __asm __volatile ("1: addl %0@+,%1;"
		      "   addxl %5,%1;"
		      "   dbf %4,1b;"
		      : "=a" (bb), "=&d" (res)
		      : "0" (bb), "1" (0), "d" (255), "d" (0));
#else
    u_long prev, *ptr;
    int i;

    res = prev = 0;
    ptr = (u_long *)bb;
    for (i = 0; i < 1024 / 4; i++) {
	res += *ptr++;
	if (res < prev)
	    res++;
	prev = res;
    }
#endif
    return(~res);
}


    /*
     *	Read the Configuration File
     */

static void ReadConfigFile(void)
{
    if (Verbose)
	printf("Reading configuration file `%s'\n", ConfigFile);

    if (!(confin = fopen(ConfigFile, "r")))
	Error_Open(ConfigFile);
    confparse();
    fclose(confin);

    Device = Config.BootDevice;
    if (!Config.Options.Default)
	Config.Options.Default = Config.Records->Label;
    if (Config.Options.Prompt)
	FixSpecialChars((char *)Config.Options.Prompt);
}


    /*
     *	Check whether two strings exist and are equal
     */

static int EqualStrings(const char *s1, const char *s2)
{
    return(s1 && s2 && !strcmp(s1, s2) ? 1 : 0);
}


    /*
     *  Interprete the special characters in a string
     */

static void FixSpecialChars(char *s)
{
    char *d = s;

    while (*s)
	if (s[0] == '\\' && s[1])
	    switch (s[1]) {
		case 'b':
		    *d++ = '\b';
		    s += 2;
		    break;
		case 'e':
		    *d++ = '\e';
		    s += 2;
		    break;
		case 'n':
		    *d++ = '\n';
		    s += 2;
		    break;
		case 'r':
		    *d++ = '\r';
		    s += 2;
		    break;
		case 't':
		    *d++ = '\t';
		    s += 2;
		    break;
		case '\\':
		    *d++ = '\\';
		    s += 2;
		    break;
		default:
		    *d++ = *s++;
		    break;
	    }
	else
	    *d++ = *s++;
    *d = '\0';
}


    /*
     *	Add a New Boot Record
     */

int AddBootRecord(const struct BootRecord *record)
{
    struct BootRecord **p, *new;

    for (p = &Config.Records; *p; p = &(*p)->Next)
	if (EqualStrings((*p)->Label, record->Label) ||
	    EqualStrings((*p)->Alias, record->Label) ||
	    EqualStrings((*p)->Label, record->Alias) ||
	    EqualStrings((*p)->Alias, record->Alias))
	    return 0;
    new = Malloc(sizeof(struct BootRecord));
    *new = *record;
    *p = new;
    return 1;
}


static dev_t ParentDeviceNr(dev_t dev_nr)
{
    switch (major(dev_nr)) {
    case LOOP_MAJOR:
    case FLOPPY_MAJOR:
	return dev_nr;
    case HD_MAJOR:
	return makedev(major(dev_nr), minor(dev_nr) & 0xc0);
    case SCSI_DISK0_MAJOR:
	return makedev(major(dev_nr), minor(dev_nr) & 0xf0);
    }
    return makedev(0, 0);
}


int AddDeviceMap(const char *path, const char *device, int unit)
{
    struct DeviceDef **p, *def;
    struct stat info;
    dev_t dev_nr;

    if (stat(path, &info))
	Error_Stat(path);
    dev_nr = ParentDeviceNr(info.st_rdev);
    for (p = &Config.Devices; *p; p = &(*p)->Next)
	if ((*p)->dev_nr == dev_nr)
	    return 0;
    def = Malloc(sizeof(*def));
    def->parent = def;
    def->dev_nr = dev_nr;
    def->path = path;
    def->device = device;
    def->unit = unit;

    *p = def;
    return 1;
}


struct DeviceDef *LookupDeviceDef(dev_t dev_nr)
{
    struct DeviceDef **p, *def, *parent;
    dev_t parent_nr;

    for (p = &Config.Devices; *p; p = &(*p)->Next)
	if ((*p)->dev_nr == dev_nr)
	    return *p;

    parent_nr = ParentDeviceNr(dev_nr);
    if (parent_nr == dev_nr)
	    Die("Unable to map device %u:%u to an amiga device\n",
		major(dev_nr), minor(dev_nr));
    parent = LookupDeviceDef(parent_nr);
    if (!parent) {
	if (BootDevice && BootDevice->dev_nr == parent_nr)
	    parent = BootDevice;
	else if (!AltDevice && Config.AltDeviceName) {
	    printf("Assuming device %u:%u resides on %s:%ld\n",
		   major(parent_nr), minor(parent_nr),
		   Config.AltDeviceName, *Config.AltDeviceUnit);
	    parent = Malloc(sizeof(*parent));
	    parent->parent = parent;
	    parent->dev_nr = parent_nr;
	    parent->device = Config.AltDeviceName;
	    parent->unit = *Config.AltDeviceUnit;
	    AltDevice = parent;
	} else
	    Die("Unable to map device %u:%u to an amiga device\n",
		major(dev_nr), minor(dev_nr));
    }
    def = Malloc(sizeof(*def));
    def->parent = parent;
    def->dev_nr = dev_nr;
    return def;
}


struct FileDef *NewFileDef(const char *path)
{
    struct FileDef *file;
    char *ptr;
    int size;

    file = Malloc(sizeof(struct FileDef));
    file->Next = NULL;
    file->Vector = NULL;
    file->Extent = NULL;

    size = strlen(path) + strlen(Root) + 1;
    file->FullPath = ptr = Malloc(size);
    strcpy(ptr, Root);
    file->Path = ptr += strlen(Root);
    strcpy(ptr, path);

    return file;
}

    /*
     *	Add a New File Definition
     */

void AddFileDef(const char *path)
{
    struct FileDef **p, *new;

    for (p = &Config.Files; *p; p = &(*p)->Next)
	if (EqualStrings((*p)->Path, path))
	    return;
    new = NewFileDef(path);
    *p = new;
}


    /*
     *	Create the Map File
     */

static void CreateMapFile(void)
{
    const struct BootRecord *record, *defrecord = NULL;
    struct FileDef *file;
    struct DeviceDef *device;
    int fd, i, vecsize;

    if (Verbose)
	printf("Creating map file `%s'\n", MapFile->Path);

    fd = open(MapFile->FullPath, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR);
    if (fd == -1)
	Error_Open(MapFile->Path);

    /* File Identification */
    WriteTagData(fd, TAG_LILO, NULL, 0);

    /* Boot Options */
    if (Verbose)
	puts("  [ Boot Options ]");
    WriteTagData(fd, TAG_HEADER, NULL, 0);
    if (Config.Options.Default)
	WriteTagString(fd, TAG_DEFAULT, Config.Options.Default);
    if (Config.Options.Auto)
	WriteTagData(fd, TAG_AUTO, Config.Options.Auto, 4);
    if (Config.Options.TimeOut)
	WriteTagData(fd, TAG_TIME_OUT, Config.Options.TimeOut, 4);
    if (Config.Options.Aux)
	WriteTagData(fd, TAG_AUX, Config.Options.Aux, 4);
    if (Config.Options.Baud)
	WriteTagData(fd, TAG_BAUD, Config.Options.Baud, 4);
    if (Config.Options.MasterPassword)
	WriteTagString(fd, TAG_MASTER_PASSWORD, Config.Options.MasterPassword);
    if (Config.Options.Debug)
	WriteTagData(fd, TAG_DEBUG, Config.Options.Debug, 4);
    if (Config.Options.Prompt)
	WriteTagString(fd, TAG_PROMPT, Config.Options.Prompt);
    if (Config.Options.KickRange)
	WriteTagData(fd, TAG_KICKRANGE, Config.Options.KickRange, 4);
    if (Config.Options.ChipSize)
	WriteTagData(fd, TAG_G_CHIP_RAM_SIZE, Config.Options.ChipSize, 4);
    for (i = 0; i < arraysize(Config.Options.FastChunks); i++)
	if (Config.Options.FastChunks[i])
	    WriteTagData(fd, TAG_G_FAST_RAM_CHUNK,
			 Config.Options.FastChunks[i], 8);
    if (Config.Options.Model)
	WriteTagData(fd, TAG_G_MODEL, Config.Options.Model, 4);
    if (Config.Options.Processor)
	WriteTagData(fd, TAG_G_PROCESSOR, Config.Options.Processor, 4);
    WriteTagData(fd, TAG_HEADER_END, NULL, 0);

    /* Boot Records */
    if (Verbose)
	puts("  [ Boot Records ]");
    else
	puts("\nBoot records:");
    for (record = Config.Records; record; record = record->Next) {
	if (EqualStrings(Config.Options.Default, record->Label) ||
	    EqualStrings(Config.Options.Default, record->Alias)) {
	    defrecord = record;
	    if (record->Password)
		Die("Default boot record `%s' must not have a password\n",
		    record->Label);
	}
	if (defrecord == record)
	    printf("\e[1m");
	printf("    %s\n", record->Label);
	if (defrecord == record)
	    printf("\e[0m");

	WriteTagString(fd, TAG_BOOT_RECORD, record->Label);
	if (record->Alias)
	    WriteTagString(fd, TAG_ALIAS, record->Alias);
	if (record->OSType)
	    WriteTagData(fd, TAG_OS_TYPE, record->OSType, 4);
	if (record->Kernel)
	    WriteTagString(fd, TAG_KERNEL, record->Kernel);
	if (record->Args)
	    WriteTagString(fd, TAG_ARGUMENTS, record->Args);
	if (record->Password)
	    WriteTagString(fd, TAG_PASSWORD, record->Password);
	if (record->ChipSize)
	    WriteTagData(fd, TAG_CHIP_RAM_SIZE, record->ChipSize, 4);
	for (i = 0; i < arraysize(record->FastChunks); i++)
	    if (record->FastChunks[i])
		WriteTagData(fd, TAG_FAST_RAM_CHUNK, record->FastChunks[i], 8);
	if (record->Model)
	    WriteTagData(fd, TAG_MODEL, record->Model, 4);
	if (record->Ramdisk)
	    WriteTagString(fd, TAG_RAMDISK, record->Ramdisk);
	if (record->Processor)
	    WriteTagData(fd, TAG_PROCESSOR, record->Processor, 4);
	WriteTagData(fd, TAG_BOOT_RECORD_END, NULL, 0);
    }
    if (!defrecord)
	Die("Default boot record `%s' does not exist\n",
	    Config.Options.Default);

    /* Device Definitions */
    if (Verbose)
	puts("  [ Device Definitions ]");
    for (device = Config.Devices; device; device = device->Next) {
	if (!device->device)
	    continue;
	WriteTagData(fd, TAG_DEVICE_DEF, &device->dev_nr, sizeof(dev_t));
	WriteTagString(fd, TAG_DEVICE_NAME, device->device);
	WriteTagData(fd, TAG_DEVICE_UNIT, &device->unit, sizeof(u_int32_t));
	WriteTagData(fd, TAG_DEVICE_DEF_END, NULL, 0);
    }

    /* File Definitions */
    if (Verbose)
	puts("  [ File Definitions ]");
    for (file = Config.Files; file; file = file->Next) {
	vecsize = 0;
	if (CreateVector(file)) {
	    if ((!defrecord->OSType || *defrecord->OSType != BOS_AMIGA) &&
		 EqualStrings(defrecord->Kernel, file->Path))
		Die("Kernel image `%s' for default boot record `%s' does not "
		    "exist\n", file->Path, defrecord->Label);
	    else
		printf("Warning: file `%s' doesn't exist\n", file->Path);
	}
	if (Verbose && file->Vector)
	    printf("    %s (%u bytes)\n", file->Path, file->Size);
	WriteTagString(fd, TAG_FILE_DEF, file->Path);
	//if (file->Vector)
	//    WriteTagData(fd, TAG_VECTOR, file->Vector, file->VectorSize);
	if (file->Extent)
	    WriteTagData(fd, TAG_FILE_EXTENT, file->Extent, file->ExtentSize);
	if (file->device && file->device->parent != BootDevice->parent)
	    WriteTagData(fd, TAG_FILE_DEVICE, &file->device->parent->dev_nr, sizeof(dev_t));
	WriteTagData(fd, TAG_FILE_DEF_END, NULL, 0);
    }

    /* End of File */
    WriteTagData(fd, TAG_EOF, NULL, 0);
    close(fd);

    if (!Verbose)
	puts("");
}


    /*
     *	Write a Tag Record to the Map File
     */

static void WriteTagData(int fd, u_long tag, const void *data, u_long size)
{
    struct TagRecord tr;

    tr.Tag = tag;
    tr.Size = size;
    if (write(fd, &tr, sizeof(tr)) != sizeof(tr))
	Error_Write(MapFile->Path);
    if (size) {
	if (write(fd, data, size) != size)
	    Error_Write(MapFile->Path);
    }
}


static void WriteTagString(int fd, u_long tag, const char *s)
{
    WriteTagData(fd, tag, s, (strlen(s) + 4) & ~3);
}


    /*
     *	Create a Vector for a File
     */

static int CreateVector(struct FileDef *file)
{
    struct DeviceDef *device;
    unsigned int offset, blksize;
    u_int64_t offset64, nextoffset;
    u_int32_t *vector;
    struct Extent *extent;
    struct stat info;
    int relblksize, blkshift, numblocks, fileblocks, sector;
    int fd, i, j;

    if (!file)
	return -1;
    fd = open(file->FullPath, O_RDONLY);
    if (fd == -1)
	return -1;
    if (fstat(fd, &info))
	Error_Stat(file->Path);
    if (!S_ISREG(info.st_mode))
	Die("%s is not a regular file\n", file->Path);

    numblocks = (info.st_size + LILO_BLOCKSIZE - 1) / LILO_BLOCKSIZE;
    file->VectorSize = sizeof(struct FileVector) + numblocks * sizeof(u_int32_t);
    file->Vector = Malloc(file->VectorSize);
    file->Vector->Size = info.st_size;
    file->ExtentSize = sizeof(struct FileExtent) + 4 * sizeof(struct Extent);
    file->Extent = Malloc(file->ExtentSize);
    file->Extent->Size = info.st_size;
    file->Extent->Count = 4;

    file->device = device = LookupDeviceDef(info.st_dev);
    Geometry(device);

    if (ioctl(fd, FIGETBSZ, &blksize) == -1)
	Error_Ioctl(file->Path, "FIGETBSZ");
    if (blksize % LILO_BLOCKSIZE)
	Die("File `%s' resides on a device with an unsupported block size "
	    "(%d bytes)\n", file->Path, blksize);
    relblksize = blksize / LILO_BLOCKSIZE;
    fileblocks = (info.st_size + blksize - 1) / blksize;
    vector = file->Vector->Vector;
    extent = file->Extent->Extent;
    for (blkshift = 0; blksize >>= 1; blkshift++)
	;
    sector = device->sectorsize;
    for (file->Extent->Shift = 0; sector >>= 1; file->Extent->Shift++)
	;
    nextoffset = 0;
    sector = 1 << (blkshift - file->Extent->Shift);

    for (i = 0; i < fileblocks; i++) {
	offset = i;
	if (ioctl(fd, FIBMAP, &offset) == -1)
	    Error_Ioctl(file->Path, "FIBMAP");
	offset64 = offset ? device->offset + ((u_int64_t)offset << blkshift) : 0;

	for (j = 0; numblocks && j < relblksize; numblocks--, j++)
	    *vector++ = offset64 + (offset64 ? j * LILO_BLOCKSIZE : 0);

	if (!i || (!extent->Block && offset64) ||
	    (extent->Block && offset64 != nextoffset)) {
	    if (i) {
		extent++;
		if (file->Extent->Extent + file->Extent->Count == extent) {
		    file->ExtentSize += 4 * sizeof(struct Extent);
		    file->Extent = realloc(file->Extent, file->ExtentSize);
		    if (!file->Extent)
			Error_NoMemory();
		    extent = file->Extent->Extent + file->Extent->Count;
		    file->Extent->Count += 4;
		}
	    }
	    extent->Block = offset64 >> file->Extent->Shift;
	    extent->Count = 0;
	}
	extent->Count += sector;
	nextoffset = offset64 ? offset64 + (1 << blkshift) : 0;
    }
    /* round down to real sector count */
    //extent->Count -= sector - ((info.st_size + sectorsize - 1) & (sector - 1));
    file->Extent->Count = ++extent - file->Extent->Extent;
    file->ExtentSize = sizeof(struct FileExtent) + file->Extent->Count * sizeof(struct Extent);
    close(fd);
    return 0;
}


static int GetGeometry(const char *devname, struct DeviceDef *device)
{
    struct hd_geometry geo;
    unsigned int tmp;
    int fd;

    fd = open(devname, O_RDONLY);
    if (fd == -1)
	return -1;

    if (ioctl(fd, BLKSSZGET, &device->sectorsize) == -1)
	Error_Ioctl(devname, "BLKSSZGET");

    if (ioctl(fd, HDIO_GETGEO, &geo) == -1)
	Error_Ioctl(devname, "HDIO_GETGEO");
    device->offset = (u_int64_t)geo.start * 512;

    if (!ioctl(fd, BLKGETSIZE64, &device->size))
	;
    else if (!ioctl(fd, BLKGETSIZE, &tmp))
	device->size = (u_int64_t)tmp * 512;
    else
	Error_Ioctl(devname, "BLKGETSIZE");

    close(fd);
    return 0;
}


    /*
     *	Get the Geometry for the Partition a File belongs to
     */

static void Geometry(struct DeviceDef *device)
{
    char devname[PATH_MAX + 1];
    struct stat info;
    DIR *dir;
    const struct dirent *dirent;

    if (device->size)
	    return;

    sprintf(devname, "/tmp/dev.%d", getpid());
    if (!mknod(devname, S_IFBLK|0600, device->dev_nr)) {
	int res = GetGeometry(devname, device);
	unlink(devname);
	if (!res)
	    return;
    }

    dir = opendir("/dev");
    if (!dir)
	Error_OpenDir("/dev");

    while ((dirent = readdir(dir))) {
	sprintf(devname, "/dev/%s", dirent->d_name);
	if (stat(devname, &info) == -1 ||
	    !S_ISBLK(info.st_mode) ||
	    info.st_rdev != device->dev_nr)
	    continue;

	if (!GetGeometry(devname, device)) {
	    closedir(dir);
	    return;
	}
    }
    Die("Can't find special device entry for device `%u:%u'\n",
	major(device->dev_nr), minor(device->dev_nr));
}


    /*
     *	Patch the Loader with the Map Vector
     */

static void PatchLoader(void)
{
    u_int32_t *data = NULL;
    u_int32_t maxsize;

    if (MapFile->device->parent != BootDevice->parent) {
	    strcpy(_MapDeviceName, MapFile->device->parent->device);
	    _MapDevice.unit = MapFile->device->parent->unit;
    }

    LoaderSize = _LoaderCodeEnd - _LoaderCodeStart;
    data = (u_int32_t *)_MapExtentData;

    maxsize = data[0] * sizeof(struct Extent);
    if (MapFile->ExtentSize > maxsize)
	Die("Map vector is too large for loader (%u bytes too much)\n",
	    MapFile->ExtentSize - maxsize);
    if (Verbose)
	printf("%d entries for map vector blocks, %u entries free\n",
	       MapFile->ExtentSize, maxsize - MapFile->ExtentSize);
    memcpy(&data[1], MapFile->Extent, MapFile->ExtentSize);
}


    /*
     *  Write the Loader (incl. Header Code)
     */

static void WriteLoader(void)
{
    int fd;

    if (Verbose)
	printf("Writing loader to file `%s'\n", LoaderFile->Path);

    fd = open(LoaderFile->FullPath, O_CREAT|O_RDWR|O_TRUNC, S_IRUSR|S_IWUSR);
    if (fd == -1)
	Error_Open(LoaderFile->Path);
    if (write(fd, _LoaderCodeStart, LoaderSize) != LoaderSize)
	Error_Write(LoaderFile->Path);
    close(fd);
}


    /*
     *	Create a Complete File Name
     */

static const char *CreateFileName(const char *name)
{
    int size;
    char *s;

    size = strlen(name) + strlen(Root) + 1;

    s = Malloc(size);
    strcpy(s, Root);
    strcat(s, name);

    return s;
}


    /*
     *  Create the File Name for the Boot Block Backup
     */

const char *CreateBackupFileName(void)
{
    int size;
    char *s;

    size = strlen(LILO_BACKUPFILEPREFIX) + 7;
    if (Root)
	size += strlen(Root);
    s = Malloc(size);
    sprintf(s, "%s%s.%02x.%02x", Root ? Root : "", LILO_BACKUPFILEPREFIX,
	    (u_int)major(BootDevice->dev_nr), (u_int)minor(BootDevice->dev_nr));
    return(s);
}


    /*
     *	Print the Usage Template and Exit
     */

static void Usage(void)
{
    Die("%s\nUsage: %s [options]\n\n"
	"Valid options are:\n"
	"    -h, --help                Display this usage information\n"
	"    -v, --verbose             Enable verbose mode\n"
	"    -f, --force               Force installation over an unknown boot block\n"
	"    -b, --backup              Force a boot block backup on installation\n"
	"    -u, --uninstall           Uninstall the boot block\n"
	"    -w, --root prefix         Use another root directory\n"
	"    -r, --restore-from file   File to restore the boot block from\n"
	"    -s, --save-to file        File to save the boot block to\n"
	"    -d, --device device       Override the block special device\n"
	"    -t, --dostype dostype     Dostype for the boot block\n"
	"    -c, --checksum            Fix the boot block checksum\n"
	"    -x, --create-reset        Create a reset boot block\n"
	"    -V, --version             Display version information\n",
	LiloVersion, ProgramName);
}


    /*
     *	Main Routine
     */

int main(int argc, char *argv[])
{
    struct stat info;
    int i;
    const char *device = NULL;

    ProgramName = argv[0];

    while (--argc > 0) {
	argv++;
	if (!strcmp(argv[0], "-h") || !strcmp(argv[0], "--help"))
	    Usage();
	else if (!strcmp(argv[0], "-v") || !strcmp(argv[0], "--verbose"))
	    Verbose = 1;
	else if (!strcmp(argv[0], "-f") || !strcmp(argv[0], "--force"))
	    Force = 1;
	else if (!strcmp(argv[0], "-b") || !strcmp(argv[0], "--backup"))
	    DoBackup = 1;
	else if (!strcmp(argv[0], "-u") || !strcmp(argv[0], "--uninstall"))
	    Uninstall = 1;
	else if (!strcmp(argv[0], "-c") || !strcmp(argv[0], "--checksum"))
	    DoChecksum = 1;
	else if (!strcmp(argv[0], "-x") || !strcmp(argv[0], "--create-reset"))
	    DoCreateReset = 1;
	else if (!strcmp(argv[0], "-V") || !strcmp(argv[0], "--version"))
	    Die(LiloVersion);
	else if (!strcmp(argv[0], "-w") || !strcmp(argv[0], "--root"))
	    if (argc-- > 1 && !Root) {
		Root = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-r") || !strcmp(argv[0], "--restore-from"))
	    if (argc-- > 1 && !RestoreBootBlock) {
		RestoreBootBlock = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-s") || !strcmp(argv[0], "--save-to"))
	    if (argc-- > 1 && !SaveBootBlock) {
		SaveBootBlock = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-d") || !strcmp(argv[0], "--device"))
	    if (argc-- > 1 && !device) {
		device = argv[1];
		argv++;
	    } else
		Usage();
	else if (!strcmp(argv[0], "-t") || !strcmp(argv[0], "--dostype"))
	    if (argc-- > 1 && !DosType) {
		DosType = argv[1];
		argv++;
	    } else
		Usage();
	else
	    Usage();
    }
    if (!Uninstall && !RestoreBootBlock && !SaveBootBlock && !DoChecksum &&
	!DoCreateReset)
        Install = 1;
    else if (DoBackup && !DoCreateReset)
	Die("You can't specify --backup with --uninstall, --restore-from, "
	    "--save-to or\n--checksum\n");
    if (RestoreBootBlock && Uninstall)
	Die("You can't specify both --restore-from and --uninstall\n");

    if (Verbose)
	printf(LiloVersion);

    ConfigFile = CreateFileName(LILO_CONFIGFILE);
    MapFile = NewFileDef(LILO_MAPFILE);
    LoaderFile = NewFileDef(LILO_LOADERFILE);
    if (SaveBootBlock)
	SaveBootBlock = CreateFileName(SaveBootBlock);
    if (RestoreBootBlock)
	RestoreBootBlock = CreateFileName(RestoreBootBlock);

    ReadConfigFile();
    if (device) {
        if (Verbose)
            printf("Using specified device `%s'\n", device);
        Device = device;
    }

    if (stat(Device, &info) == -1)
	Error_Stat(Device);
    if (!S_ISBLK(info.st_mode))
	Die("%s is not a block special device\n", Device);
    AddDeviceMap(Device, NULL, 0);
    BootDevice = LookupDeviceDef(info.st_rdev);
    BackupFile = CreateBackupFileName();

    ReadBootBlock(Device);
    printf("%s: ", Device);
    DosTypeID = BootBlock.ID;
    for (i = 0; DosTypeEntries[i].ID; i++)
	if (BootBlock.ID == DosTypeEntries[i].ID)
	    break;
    if (DosTypeEntries[i].ID)
	printf("%s", DosTypeEntries[i].Description);
    else {
	printf("Type 0x%08lx", BootBlock.ID);
	if (!Force) {
	    printf("\n");
	    Die("Use option --force to force continue\n");
	}
	DosTypeID = ID_BOOT_DISK;
    }
    printf("%s, Boot block checksum is %s\n",
	   BootBlock.LiloID == LILO_ID ? " (created by Lilo)" : "",
	   CalcChecksum() ? "Incorrect" : "OK");

    if (SaveBootBlock) {
	WriteFile(SaveBootBlock, &BootBlock, sizeof(BootBlock));
	printf("Boot block saved to file `%s'\n", SaveBootBlock);
    }
    if (RestoreBootBlock) {
	ReadBootBlock(RestoreBootBlock);
	printf("Boot block read from file `%s'\n", RestoreBootBlock);
    }
    if (Uninstall) {
	ReadBootBlock(BackupFile);
	printf("Original boot block read from file `%s'\n", BackupFile);
    }
    if (DosType)
	SetDosType();
    if (Install) {
	BackupBootBlock();
	CreateMapFile();
	if (CreateVector(MapFile))
	    Error_Open(MapFile->Path);
	PatchLoader();
	WriteLoader();
	if (CreateVector(LoaderFile))
	    Error_Open(LoaderFile->Path);
	CreateBootBlock(0);
    }
    if (DoCreateReset) {
	BackupBootBlock();
	CreateBootBlock(1);
    }
    if (DoChecksum)
	FixChecksum();
    if (RestoreBootBlock || Install || Uninstall || DoChecksum || DoCreateReset)
	WriteBootBlock(Device);
    exit(0);
}
