/*
 *  Linux/m68k Elf boot linker - creates an executable from an object file
 *
 *  (C) Copyright 1998 by Roman Zippel
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License.  See the file COPYING for more details.
 *
 * $Id$
 *
 */

#define _GNU_SOURCE

#include <sys/stat.h>
#include <sys/mman.h>
#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <endian.h>
#include <byteswap.h>

#define DEBUG		0
#define NEED_SWAP	(__BYTE_ORDER != __BIG_ENDIAN)

static int ll_obj_fd = -1;
static int ll_do_source;
static int ll_do_relocate;

static Elf32_Ehdr ll_ehdr;
static Elf32_Shdr *ll_shdr;

static void *(*ll_sections);
static char *ll_strings;
static int ll_bss;
static unsigned int ll_alloc_size;
static unsigned int ll_load_size;
static unsigned int ll_reloc_cnt;

static Elf32_Sym *ll_symtab;
static int ll_sym_idx;
static char *ll_sym_str;

#define LL_HEADER_SIZE	sizeof(struct ll_header_s)
struct ll_header_s {
	unsigned int prog_entry;
	unsigned int prog_size;
	unsigned int reloc_cnt;
};
static struct ll_header_s ll_header;
static u_int8_t *ll_exec_mem;

#if NEED_SWAP
#define ll_swap16(val)	bswap_16(val)
#define ll_swap32(val)	bswap_32(val)
#else
#define ll_swap16(val)	(val)
#define ll_swap32(val)	(val)
#endif

#if DEBUG
#define printd(fmt...)	printf(fmt)
#else
#define printd(fmt...)
#endif

static int ll_exit(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(1);
}

static ssize_t ll_pread(int fd, void *buf, size_t size, off_t offset)
{
	ssize_t read, res;

	read = 0;
	do {
		res = pread(fd, buf, size, offset);
		if (res <= 0) {
			if (res == 0)
				break;
			if (errno == EINTR || errno == EAGAIN)
				continue;
			if (!read)
				read = -1;
			break;
		}
		buf = (void *)((unsigned long)buf + res);
		size -= res;
		offset += res;
		read += res;
	} while (size > 0);
	return read;
}

static ssize_t ll_pwrite(int fd, const void *buf, size_t size, off_t offset)
{
	ssize_t write, res;

	write = 0;
	do {
		res = pwrite(fd, buf, size, offset);
		if (res < 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			if (!res)
				write = -1;
			break;
		}
		buf = (const void *)((unsigned long)buf + res);
		size -= res;
		offset += res;
		write += res;
	} while (size > 0);
	return write;
}

static void ll_open_object(char *file)
{
	ll_obj_fd = open(file, O_RDONLY);
	if (ll_obj_fd == -1)
		ll_exit("Can't open loader object (%s)!\n", strerror(errno));

#if 0
	if (fstat(ll_obj_fd, &stat) == -1)
		ll_exit("Can't stat loader object (%s)!\n", strerror(errno));

	lptr = mmap(NULL, stat.st_size, PROT_READ, MAP_SHARED, lfd, 0);
	if (lptr == (char *)-1) {
		printf("Can't map loader object (%s)!\n",
				strerror(errno));
		return 1;
	}
#endif
}

static void ll_load_ehdr(void)
{
	if (ll_pread(ll_obj_fd, &ll_ehdr, sizeof(ll_ehdr), 0) != sizeof(ll_ehdr))
		ll_exit("Unable to read ELF header (%s)!\n", strerror(errno));

	if (NEED_SWAP) {
		ll_ehdr.e_type      = ll_swap16(ll_ehdr.e_type);
		ll_ehdr.e_machine   = ll_swap16(ll_ehdr.e_machine);
		ll_ehdr.e_version   = ll_swap32(ll_ehdr.e_version);
		ll_ehdr.e_entry     = ll_swap32(ll_ehdr.e_entry);
		ll_ehdr.e_phoff     = ll_swap32(ll_ehdr.e_phoff);
		ll_ehdr.e_shoff     = ll_swap32(ll_ehdr.e_shoff);
		ll_ehdr.e_flags     = ll_swap32(ll_ehdr.e_flags);
		ll_ehdr.e_ehsize    = ll_swap16(ll_ehdr.e_ehsize);
		ll_ehdr.e_phentsize = ll_swap16(ll_ehdr.e_phentsize);
		ll_ehdr.e_phnum     = ll_swap16(ll_ehdr.e_phnum);
		ll_ehdr.e_shentsize = ll_swap16(ll_ehdr.e_shentsize);
		ll_ehdr.e_shnum     = ll_swap16(ll_ehdr.e_shnum);
		ll_ehdr.e_shstrndx  = ll_swap16(ll_ehdr.e_shstrndx);
	}

	if (!memcmp(ll_ehdr.e_ident, ELFMAG, SELFMAG) &&
		    ll_ehdr.e_type == ET_REL &&
		    ll_ehdr.e_machine == EM_68K &&
		    ll_ehdr.e_version == EV_CURRENT)
		printf("ELF/m68k relocatable executable found.\n");
	else
		ll_exit("No ELF/m68k relocatable executable found!\n");

	ll_sections = malloc(ll_ehdr.e_shnum * sizeof(*ll_sections));
}

static void ll_load_shdr(void)
{
	int size;

	if (!ll_ehdr.e_shoff)
		ll_exit("Missing section header table in object!\n");

	size = ll_ehdr.e_shnum * sizeof(Elf32_Shdr);
	ll_shdr = malloc(size);
	if (!ll_shdr)
		ll_exit("Can't alloc mem for section header table!\n");

	if (ll_pread(ll_obj_fd, ll_shdr, size, ll_ehdr.e_shoff) != size)
		ll_exit("Unable to read section header table (%s)!\n", strerror(errno));

	size = ll_ehdr.e_shnum * sizeof(void *);
	ll_sections = malloc(size);
	if (!ll_sections)
		ll_exit("Can't alloc mem for section table!\n");
	memset(ll_sections, 0, size);

	if (NEED_SWAP) {
		Elf32_Shdr *shdr;
		int sidx;

		for (shdr = ll_shdr, sidx = 0; sidx < ll_ehdr.e_shnum; shdr++, sidx++) {
			shdr->sh_name      = ll_swap32(shdr->sh_name);
			shdr->sh_type      = ll_swap32(shdr->sh_type);
			shdr->sh_flags     = ll_swap32(shdr->sh_flags);
			shdr->sh_addr      = ll_swap32(shdr->sh_addr);
			shdr->sh_offset    = ll_swap32(shdr->sh_offset);
			shdr->sh_size      = ll_swap32(shdr->sh_size);
			shdr->sh_link      = ll_swap32(shdr->sh_link);
			shdr->sh_info      = ll_swap32(shdr->sh_info);
			shdr->sh_addralign = ll_swap32(shdr->sh_addralign);
			shdr->sh_entsize   = ll_swap32(shdr->sh_entsize);
		}
	}
}

static void *ll_open_section(int idx)
{       
	void *ptr;
	int size;

	ptr = ll_sections[idx];
	if (ptr)
		return ptr;

	size = ll_shdr[idx].sh_size;
	ptr = malloc(size);
	if (!ptr)
		ll_exit("Can't alloc mem for section %d!\n", idx);

	if (ll_pread(ll_obj_fd, ptr, size, ll_shdr[idx].sh_offset) != size)
		ll_exit("Unable to read section %d (%s)!\n", strerror(errno));
	ll_sections[idx] = ptr;

	return ptr;
}

static void *ll_get_section(int idx)
{
	return ll_sections[idx];
}

static void ll_init_sizes(void)
{
	Elf32_Shdr *shdr;
	int sidx;

	ll_bss        = SHN_UNDEF;
	ll_alloc_size = ll_do_relocate ? LL_HEADER_SIZE : 0;
	ll_reloc_cnt  = 0;

	for (shdr = ll_shdr, sidx = 0; sidx < ll_ehdr.e_shnum; shdr++, sidx++) {
		if (shdr->sh_type == SHT_NULL)
			continue;

		if (shdr->sh_flags & SHF_ALLOC) {
			if (shdr->sh_type == SHT_PROGBITS) {
				int align = shdr->sh_addralign;
				if (align--)
					ll_alloc_size = (ll_alloc_size + align) & ~align;
				shdr->sh_addr = ll_alloc_size;
				ll_alloc_size += shdr->sh_size;
				printd("Load section '%s' to offset 0x%x\n",
				       ll_strings + shdr->sh_name, shdr->sh_addr);
			} else if (shdr->sh_type == SHT_NOBITS) {
				if (ll_bss != SHN_UNDEF)
					ll_exit("Found more than one bss section!\n");

				printd("Defining bss section '%s' later\n",
						ll_strings + shdr->sh_name);
				ll_bss = sidx;
			} else
				ll_exit("Can't alloc section '%s' of type %d\n!",
				       ll_strings[shdr->sh_name], shdr->sh_type);
		}

		if (shdr->sh_type == SHT_REL) {
#if 0
			ll_reloc_cnt += shdr->sh_size / sizeof(Elf32_Rel);
#endif
			ll_exit("Can't handle SHT_REL section (yet)!\n");
		}

		if (ll_do_relocate && shdr->sh_type == SHT_RELA)
			ll_reloc_cnt += shdr->sh_size / sizeof(Elf32_Rela);
	}

	if (ll_bss == SHN_UNDEF)
		ll_exit("Found no bss section!\n");

	ll_load_size = ll_alloc_size;
}

static Elf32_Sym *ll_get_symtab(int sidx)
{
	Elf32_Sym *sym_tab = ll_get_section(sidx);
	Elf32_Sym *sym;
	int idx;

	if (sym_tab)
		return sym_tab;

	sym_tab = ll_open_section(sidx);
	if (!NEED_SWAP)
		return sym_tab;

	sym = sym_tab;
	idx = ll_shdr[sidx].sh_size / sizeof(Elf32_Sym);
	for (; idx > 0; sym++, idx--) {
		sym->st_name  = ll_swap32(sym->st_name);
		sym->st_value = ll_swap32(sym->st_value);
		sym->st_size  = ll_swap32(sym->st_size);
		sym->st_shndx = ll_swap16(sym->st_shndx);
	}

	return sym_tab;
}

static Elf32_Rela *ll_get_rela(int sidx)
{
	Elf32_Rela *rela_tab = ll_get_section(sidx);
	Elf32_Rela *rela;
	int idx;

	if (rela_tab)
		return rela_tab;
	
	rela_tab = ll_open_section(sidx);
	if (!NEED_SWAP)
		return rela_tab;

	rela = rela_tab;
	idx = ll_shdr[sidx].sh_size / sizeof(Elf32_Rela);

	for (; idx > 0; rela++, idx--) {
		rela->r_offset = ll_swap32(rela->r_offset);
		rela->r_info   = ll_swap32(rela->r_info);
		rela->r_addend = ll_swap32(rela->r_addend);
	}

	return rela_tab;
}

static void ll_load_symbol_table(void)
{
	Elf32_Shdr *shdr;
	int sidx;

	ll_symtab = NULL;

	for (shdr = ll_shdr, sidx = 0; sidx < ll_ehdr.e_shnum; shdr++, sidx++) {
		if (shdr->sh_type != SHT_SYMTAB)
			continue;

		if (ll_symtab)
			ll_exit("Found more than one symbol section!\n");

		ll_symtab = ll_get_symtab(sidx);

		ll_sym_idx = sidx;
		ll_sym_str = ll_open_section(shdr->sh_link);
	}

	if (!ll_symtab)
		ll_exit("Found no symbol section!\n");
}

static void ll_alloc_bss_symbols(void)
{
	Elf32_Sym *sym;
	int idx, align;

	align = ll_shdr[ll_bss].sh_addralign;
	if (align--)
		ll_alloc_size = (ll_alloc_size + align) & ~align;
	ll_shdr[ll_bss].sh_addr = ll_alloc_size;
	ll_alloc_size += ll_shdr[ll_bss].sh_size;

	for (sym = ll_symtab, idx = ll_shdr[ll_sym_idx].sh_size / sizeof(Elf32_Sym);
	     idx > 0; sym++, idx--) {
		if (sym->st_shndx == SHN_COMMON) {
			sym->st_shndx = ll_bss;
			align = sym->st_value;
			if (align--)
				ll_alloc_size = (ll_alloc_size + align) & ~align;
			sym->st_value = ll_alloc_size - ll_shdr[ll_bss].sh_addr;
			ll_alloc_size += sym->st_size;
			printd("Allocating '%s' at %x\n",
			       &ll_sym_str[sym->st_name], sym->st_value);
		}
	}
	ll_shdr[ll_bss].sh_size = ll_alloc_size - ll_shdr[ll_bss].sh_addr;

	printd("Load bss section to offset 0x%x size 0x%x\n", ll_shdr[ll_bss].sh_addr, ll_shdr[ll_bss].sh_size);
	printd("Program size is 0x%x\n", ll_alloc_size);
}

static void ll_resolve_symbols(void)
{
	Elf32_Sym *sym;
	char *name;
	int idx;

	sym = ll_symtab + 1;
	idx = ll_shdr[ll_sym_idx].sh_size / sizeof(Elf32_Sym) - 1;
	for (; idx > 0; sym++, idx--) {
		name = ll_sym_str + sym->st_name;
		if (sym->st_shndx == SHN_UNDEF) {
			if (!strcmp(name, "_edata")) {
				sym->st_value = ll_shdr[ll_bss].sh_size;
				sym->st_size  = 0;
				sym->st_shndx = ll_bss;
				sym->st_info  = ELF32_ST_INFO(STB_GLOBAL, STT_OBJECT);
				printd("Allocating '%s' at %x\n", name, sym->st_value);
			} else if (!strcmp(name, "__end")) {
				sym->st_value = ll_shdr[ll_bss].sh_size;
				sym->st_size  = 0;
				sym->st_shndx = ll_bss;
				sym->st_info  = ELF32_ST_INFO(STB_GLOBAL, STT_OBJECT);
				printd("Allocating '%s' at %x:%x\n", name,
				       ll_shdr[sym->st_shndx].sh_addr, sym->st_value);
			} else if (!strcmp(name, "__bss_start")) {
				sym->st_value = 0;
				sym->st_size  = 0;
				sym->st_shndx = ll_bss;
				sym->st_info  = ELF32_ST_INFO(STB_GLOBAL, STT_OBJECT);
				printd("Allocating '%s' at %x:%x\n", name,
				       ll_shdr[sym->st_shndx].sh_addr, sym->st_value);
			} else
				ll_exit("Can't resolve symbol '%s'!\n", name);
		}
		if (!strcmp(name, "_start"))
			ll_header.prog_entry = ll_swap32(ll_shdr[sym->st_shndx].sh_addr + sym->st_value);
	}
}

void ll_open_exec(void)
{
	ll_exec_mem = malloc(ll_load_size + (ll_reloc_cnt + 1) * sizeof(u_int32_t));
	if (!ll_exec_mem)
		ll_exit("Can't alloc mem for relocate data!\n");
	memset(ll_exec_mem, 0, ll_load_size);
}

void ll_create_exec(char *out_name)
{
	Elf32_Shdr *shdr, *prog_shdr, *sym_shdr;
	Elf32_Sym *sym_table, *sym;
	Elf32_Rela *rela_table, *rela;
	unsigned char *prog_data;
	void *dataptr;
	char *sym_str;
	int sidx, ridx, reloc_cnt;
	u_int32_t *reloc_ptr;
	unsigned int val = 0;

	reloc_ptr = (u_int32_t *)(ll_exec_mem + ll_load_size);
	ll_reloc_cnt = 0;
 
	for (shdr = ll_shdr, sidx = 0; sidx < ll_ehdr.e_shnum; shdr++, sidx++) {
		if (shdr->sh_type != SHT_RELA)
			continue;

		prog_shdr = ll_shdr + shdr->sh_info;
		if (prog_shdr->sh_type != SHT_PROGBITS)
			ll_exit("Invalid data section '%s' in relocation info!\n",
				ll_strings + prog_shdr->sh_name);

		if (!(prog_shdr->sh_flags & SHF_ALLOC)) {
			printd("Skipping relocating '%s' with '%s'\n",
			       ll_strings + prog_shdr->sh_name, ll_strings + shdr->sh_name);
			continue;
		}

		sym_shdr = ll_shdr + shdr->sh_link;
		if (sym_shdr->sh_type != SHT_SYMTAB)
			ll_exit("Invalid symbol section '%s' in relocation info!\n",
				ll_strings + sym_shdr->sh_name);

		rela_table = ll_get_rela(sidx);
		prog_data = ll_open_section(shdr->sh_info);
		sym_table = ll_get_symtab(shdr->sh_link);
		sym_str = ll_open_section(ll_shdr[shdr->sh_link].sh_link);

		printd("Use '%s' to relocate '%s'\n",
		       ll_strings + shdr->sh_name, ll_strings + prog_shdr->sh_name);

		rela = rela_table;
		reloc_cnt = shdr->sh_size / sizeof(Elf32_Rela);

		for (ridx = 0; ridx < reloc_cnt; rela++, ridx++) {
			sym = &sym_table[ELF32_R_SYM(rela->r_info)];
			printd("relocate %x with %s add %x type %x\n", rela->r_offset,
			       sym_str + sym->st_name, rela->r_addend, ELF32_R_TYPE(rela->r_info));
			switch (ELF32_ST_TYPE(sym->st_info)) {
			case STT_NOTYPE:
			case STT_OBJECT:
			case STT_FUNC:
			case STT_SECTION:
				if (sym->st_shndx == SHN_UNDEF)
					ll_exit("Invalid symbol to resolved!\n");
				val = sym->st_shndx == SHN_ABS ? 0 : ll_shdr[sym->st_shndx].sh_addr;
				val += sym->st_value;
				break;
			default:
				ll_exit("Don't know how to resolve symbol '%s' of type %d (yet)\n",
					sym_str + sym->st_name, ELF32_ST_TYPE(sym->st_size));
			}

			dataptr = prog_data + rela->r_offset;
			switch (ELF32_R_TYPE(rela->r_info)) {
			case R_68K_32:
				if (*(unsigned int *)dataptr)
					printf("Warning: relocation storage at 0x%x isn't empty!\n",
					       rela->r_offset);

				*(unsigned int *)dataptr = ll_swap32(val + rela->r_addend);

				if (ll_do_relocate) {
					*reloc_ptr++ = ll_swap32(prog_shdr->sh_addr + rela->r_offset);
					ll_reloc_cnt++;
				}
				break;
			case R_68K_PC32:
				if (*(unsigned int *)dataptr)
					printf("Warning: relocation storage at 0x%x isn't empty!\n",
					       rela->r_offset);

				*(unsigned int *)dataptr = ll_swap32(val + rela->r_addend -
								     (prog_shdr->sh_addr + rela->r_offset));
				break;
			default:
				ll_exit("Don't know how to relocate type %d (yet)!\n",
					ELF32_R_TYPE(rela->r_info));
			}
		}

		memcpy(ll_exec_mem + prog_shdr->sh_addr, prog_data, prog_shdr->sh_size);
		prog_shdr->sh_flags &= ~SHF_ALLOC;

		free(prog_data);
	}

	for (shdr = ll_shdr, sidx = 0; sidx < ll_ehdr.e_shnum; shdr++, sidx++) {
		if (shdr->sh_type != SHT_PROGBITS || !(shdr->sh_flags & SHF_ALLOC))
			continue;
		printd("Writing '%s' without relocations\n", ll_strings + shdr->sh_name);
		prog_data = ll_open_section(sidx);

		memcpy(ll_exec_mem + shdr->sh_addr, prog_data, shdr->sh_size);

		free(prog_data);
	}

	ll_header.prog_size   = ll_swap32(ll_load_size);
	ll_header.reloc_cnt   = ll_swap32(ll_reloc_cnt);
	if (ll_do_relocate)
		memcpy(ll_exec_mem, &ll_header, sizeof(ll_header));

	if (ll_do_source) {
		FILE *file;
		Elf32_Sym *sym;
		unsigned int i, size;

		file = fopen(out_name, "w");
		if (!file)
			ll_exit("Can't create loader (%s)!\n", strerror(errno));
		size = ll_shdr[ll_sym_idx].sh_size / sizeof(Elf32_Sym);
		sym = ll_symtab + 1;
		for (i = 1; i < size; sym++, i++) {
			if (ELF32_ST_BIND(sym->st_info) == STB_GLOBAL)
				fprintf(file, "%s = . + %u\n", ll_sym_str + sym->st_name,
					ll_shdr[sym->st_shndx].sh_addr + sym->st_value);
		}
		size = ll_load_size + ll_reloc_cnt * sizeof(u_int32_t);
		for (i = 0; i < size; i++) {
			if ((i % 16) == 0)
				fprintf(file, "\t.byte\t0x%02x", ll_exec_mem[i]);
			else
				fprintf(file, ",0x%02x", ll_exec_mem[i]);
			if ((i % 16) == 15)
				fprintf(file, "\n");
		}
		if (ll_alloc_size > size) {
			for (; i < ll_alloc_size; i++) {
				if ((i % 16) == 0)
					fprintf(file, "\t.byte\t0x00");
				else
					fprintf(file, ",0x00");
				if ((i % 16) == 15)
					fprintf(file, "\n");
			}
		}
		fprintf(file, "\n");
	} else {
		int fd;

		fd = creat(out_name, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		if (fd == -1)
			ll_exit("Can't create loader (%s)!\n", strerror(errno));

		ll_pwrite(fd, ll_exec_mem, ll_load_size + ll_reloc_cnt * sizeof(u_int32_t), 0);

		if (ll_alloc_size > ll_load_size + ll_reloc_cnt * sizeof(unsigned int)) {
			printd("Extending loader to full size\n");
			if (ftruncate(fd, ll_alloc_size) == -1)
				ll_exit("Can't extend loader till %d (%s)!\n",
					ll_alloc_size, strerror(errno));
		}
		close(fd);
	}
}

int main(int argc, char **argv)
{
	char *progname = argv[0];

	argc--;
	argv++;
	while (argc > 0 && argv[0][0] == '-') {
		switch (argv[0][1]) {
		case 'S':
			ll_do_source = 1;
			break;
		case 'R':
			ll_do_relocate = 1;
			break;
		default:
			ll_exit("Unknown option '%c'\n", argv[0][1]);
		}
		argc--;
		argv++;
	}
	if (argc != 2)
		ll_exit("Usage: %s [-S] [-R] <loader.o> <loader>\n", progname);

	ll_open_object(argv[0]);

	ll_load_ehdr();

	ll_load_shdr();

	ll_strings = ll_open_section(ll_ehdr.e_shstrndx);

	ll_init_sizes();

	ll_load_symbol_table();

	ll_alloc_bss_symbols();

	ll_resolve_symbols();

	ll_open_exec();

	ll_create_exec(argv[1]);

	free(ll_exec_mem);

	close(ll_obj_fd);

	printf("Loader written.\n");

	return 0;
}
